import 'dart:async';
import 'package:http/http.dart' as ht;
import 'dart:convert' as codec;
import 'expect.dart';
import 'response.dart';
import 'package:http_parser/http_parser.dart';
import 'package:client_cookie/client_cookie.dart';

class CookieJar {
  final CookieStore store = new CookieStore();
  void cookieStoreInterceptor(Route r) {
    r.cookies(store.cookies);
    r.interceptAfter((resp) {
      store.addFromHeader(resp.headers['set-cookie']);
    });
  }
}

typedef void Before(Route route);

typedef void After(Response response);

Route route(String authorityOrPath, [String path]) =>
    new Route(authorityOrPath, path);

Get get(String authorityOrPath, [String path]) =>
    new Get(authorityOrPath, path);

Post post(String authorityOrPath, [String path]) =>
    new Post(authorityOrPath, path);

Put put(String authorityOrPath, [String path]) =>
    new Put(authorityOrPath, path);

Delete delete(String authorityOrPath, [String path]) =>
    new Delete(authorityOrPath, path);

ht.BaseClient globalClient;

/// Build fluent REST APIs routes
///
/// Example:
///     route('/book')
///       .query('count', '10')
///       .get.fetchList((m) => new Book.fromMap(m));
class Route {
  String _path = '';

  String _authority;

  Map<String, String> _query = {};

  Map<String, String> _headers = {};

  final List<ClientCookie> _cookies = new List<ClientCookie>();

  final List<Before> before = new List<Before>();
  final List<After> after = <After>[];

  ht.BaseClient _client;

  /// Construct [Route] object with [path]
  Route([String authorityOrPath, String path]) {
    if (authorityOrPath != null) {
      if (path != null) {
        authority(authorityOrPath, path);
      } else
        this.path(authorityOrPath);
    }
  }

  /// Set the [client] used to make HTTP requests
  Route withClient(ht.BaseClient client) {
    _client = client;
    return this;
  }

  Route http(String authority, [String path]) {
    _authority = 'http://${authority}';
    if (path != null) this.path(path);
    return this;
  }

  Route https(String authority, [String path]) {
    _authority = 'https://${authority}';
    if (path != null) this.path(path);
    return this;
  }

  /// Set authority of the URL
  Route authority(String authority, [String path]) {
    _authority = authority;
    if (path != null) this.path(path);
    return this;
  }

  /// Append path segments to the URL
  Route path(String path) {
    if (path.isEmpty) return this;
    if (!_path.endsWith('/') && !path.startsWith('/')) _path += '/';
    _path += path;
    return this;
  }

  /// Add query parameters
  Route query(String key, value) {
    _query[key] = value.toString();
    return this;
  }

  /// Add query parameters
  Route queries(Map<String, dynamic> value) {
    _query.addAll(value);
    return this;
  }

  /// Add headers
  Route header(String key, String value) {
    _headers[key] = value;
    return this;
  }

  /// Add headers
  Route headers(Map<String, String> values) {
    _headers.addAll(values);
    return this;
  }

  Route cookie(ClientCookie cookie) {
    _cookies.add(cookie);
    return this;
  }

  Route cookies(List<ClientCookie> cookie) {
    _cookies.addAll(cookie);
    return this;
  }

  Route interceptBefore(Before interceptor) {
    before.add(interceptor);
    return this;
  }

  Route interceptAfter(After interceptor) {
    after.add(interceptor);
    return this;
  }

  /// URL
  String get url {
    String path = _path.split('/').where((s) => s.isNotEmpty).join('/');
    if (path.isNotEmpty) {
      if (_path.endsWith('/')) path += '/';
    }

    if (_authority == null && _query == null) {
      return path;
    }
    StringBuffer sb = new StringBuffer();
    if (_authority != null) sb.write(_authority);
    if (_authority == null || !_authority.endsWith('/')) sb.write('/');
    sb.write(path);
    if (_query == null) {
      return sb.toString();
    }
    _makeQueryParams(sb, _query);
    return sb.toString();
  }

  /// Converts to [Get] requester
  Get get get => new Get.copy(this);

  /// Converts to [Post] requester
  Post get post => new Post.copy(this);

  /// Converts to [Put] requester
  Put get put => new Put.copy(this);

  /// Converts to [Delete] requester
  Delete get delete => new Delete.copy(this);

  static void _makeQueryParams(StringBuffer sb, Map<String, dynamic> query) {
    if (query.length == 0) return;
    sb.write('?');

    void writeQuery(String key, String value, [bool isFirst = false]) {
      if (!isFirst) sb.write('&');
      sb.write(Uri.encodeQueryComponent(key));
      if (value != null && value.isNotEmpty) {
        sb.write("=");
        sb.write(Uri.encodeQueryComponent(value));
      }
    }

    void writeQueries(String key, value, [bool isFirst = false]) {
      if (value == null || value is String) {
        writeQuery(key, value, isFirst);
        return;
      }
      if (value is Iterable<String>) {
        for (int i = 0; i < value.length; i++) {
          if (i == 0) {
            writeQuery(key, value.elementAt(i), isFirst);
          } else {
            writeQuery(key, value.elementAt(i), false);
          }
        }
        return;
      }
    }

    for (int i = 0; i < query.length; i++) {
      String key = query.keys.elementAt(i);
      if (i == 0) {
        writeQueries(key, query[key], true);
      } else {
        writeQueries(key, query[key], false);
      }
    }
  }
}

/// Build fluent REST GET APIs
///
/// Example:
///     get('/book')
///       .query('count', '10')
///       .fetchList((m) => new Book.fromMap(m));
class Get extends Route {
  Get([String authorityOrPath, String path]) : super(authorityOrPath, path);

  Get.copy(Route route) {
    _authority = route._authority;
    _path = route._path;
    _query = route._query;
    _headers = route._headers;
    _client = route._client;
    before.addAll(route.before);
    after.addAll(route.after.toList());
  }

  Get withClient(ht.BaseClient client) => super.withClient(client);

  Get http(String authority, [String path]) => super.http(authority, path);

  Get https(String authority, [String path]) => super.https(authority, path);

  Get authority(String authority, [String path]) =>
      super.authority(authority, path);

  Get path(String path) => super.path(path);

  Get query(String key, value) => super.query(key, value);

  Get queries(Map<String, dynamic> value) => super.queries(value);

  Get header(String key, String value) => super.header(key, value);

  Get headers(Map<String, String> values) => super.headers(values);

  Get cookie(ClientCookie cookie) => super.cookie(cookie);

  Get cookies(List<ClientCookie> cookies) => super.cookies(cookies);

  Get interceptBefore(Before interceptor) => super.interceptBefore(interceptor);

  Get interceptAfter(After interceptor) => super.interceptAfter(interceptor);

  AsyncStringResponse get go {
    for (Before mod in before) mod(this);
    if (_cookies.length != 0)
      _headers['Cookie'] = ClientCookie.toHeader(_cookies);
    return new AsyncStringResponse.from(
        (_client ?? globalClient).get(url, headers: _headers))
      ..then((r) {
        for (After rp in after) rp(r);
      });
  }

  Future<T> one<T, F>([T convert(F d)]) async {
    StringResponse resp = await go;
    if (resp.statusCode == 200) return go.decode(convert);
    throw resp;
  }

  Future<List<T>> list<T, F>([T convert(F d)]) async {
    StringResponse resp = await go;
    if (resp.statusCode == 200) return go.decodeList(convert);
    throw resp;
  }

  AsyncStringResponse expect(List<Checker<Response>> conditions) =>
      go..expect(conditions);

  AsyncStringResponse exact(
          {int statusCode,
          String body,
          List<int> bytes,
          String mimeType,
          String encoding,
          Map<String, String> headers,
          int contentLength}) =>
      go
        ..exact(
            statusCode: statusCode,
            body: body,
            bytes: bytes,
            mimeType: mimeType,
            encoding: encoding,
            headers: headers,
            contentLength: contentLength);
}

/// Build fluent REST POST APIs
///
/// Example:
///     post('/book/${id}')
///       .json(book.toMap)
///       .fetch((m) => new Book.fromMap(m));
class Post extends Route {
  dynamic _body;

  Post([String authorityOrPath, String path]) : super(authorityOrPath, path);

  Post.copy(Route route) {
    _authority = route._authority;
    _path = route._path;
    _query = route._query;
    _headers = route._headers;
    _client = route._client;
    before.addAll(route.before);
    after.addAll(route.after.toList());
  }

  Post withClient(ht.BaseClient client) => super.withClient(client);

  Post http(String authority, [String path]) => super.http(authority, path);

  Post https(String authority, [String path]) => super.https(authority, path);

  Post authority(String authority, [String path]) =>
      super.authority(authority, path);

  Post path(String path) => super.path(path);

  Post query(String key, value) => super.query(key, value);

  Post queries(Map<String, dynamic> value) => super.queries(value);

  Post header(String key, String value) => super.header(key, value);

  Post headers(Map<String, String> values) => super.headers(values);

  Post cookie(ClientCookie cookie) => super.cookie(cookie);

  Post cookies(List<ClientCookie> cookies) => super.cookies(cookies);

  Post body(String body) {
    _body = body;
    return this;
  }

  Post json(body, {bool setHeaders: true}) {
    _body = codec.json.encode(body);
    if (setHeaders) {
      header('content-type', 'application/json');
      header('Accept', 'application/json');
    }
    return this;
  }

  Post multipart(Map<String, String> values) {
    if (_body is! Map<String, Multipart>) {
      _body = <String, Multipart>{};
    }
    for (String field in values.keys) {
      _body[field] = new _MultipartString(values[field]);
    }
    return this;
  }

  Post multipartFile(String field, List<int> value,
      {String filename, MediaType contentType}) {
    if (_body is! Map<String, Multipart>) {
      _body = <String, Multipart>{};
    }
    _body[field] =
        new _MultipartFile(value, filename: filename, contentType: contentType);
    return this;
  }

  Post multipartStringFile(String field, String value,
      {String filename, MediaType contentType}) {
    if (_body is! Map<String, Multipart>) {
      _body = <String, Multipart>{};
    }
    _body[field] = new _MultipartStringFile(value,
        filename: filename, contentType: contentType);
    return this;
  }

  Post urlEncodedForm(Map<String, String> values) {
    if (_body is! Map<String, String>) {
      _body = <String, String>{};
    }
    _body.addAll(values);
    return this;
  }

  AsyncStringResponse get go {
    for (Before mod in before) mod(this);
    if (_cookies.length != 0)
      _headers['Cookie'] = ClientCookie.toHeader(_cookies);
    if (_body is String || _body is Map<String, String> || _body == null) {
      return new AsyncStringResponse.from(
          (_client ?? globalClient).post(url, headers: _headers, body: _body));
    } else if (_body is Map<String, Multipart>) {
      final body = _body as Map<String, Multipart>;
      final r = new ht.MultipartRequest('POST', Uri.parse(url));
      for (final String field in body.keys) {
        final Multipart value = body[field];
        if (value is _MultipartString) {
          r.fields[field] = value.value;
        } else if (value is _MultipartStringFile) {
          r.files.add(new ht.MultipartFile.fromString(field, value.value,
              filename: value.filename, contentType: value.contentType));
        } else if (value is _MultipartFile) {
          r.files.add(new ht.MultipartFile.fromBytes(field, value.value,
              filename: value.filename, contentType: value.contentType));
        }
      }
      r.headers.addAll(_headers);
      return new AsyncStringResponse.from(
          r.send().then((r) => ht.Response.fromStream(r)))
        ..then((r) {
          for (After rp in after) rp(r);
        });
    } else {
      throw new Exception('Invalid body!');
    }
  }

  Future<T> one<T, F>([T convert(F d)]) async {
    StringResponse resp = await go;
    if (resp.statusCode == 200) return go.decode(convert);
    throw resp;
  }

  Future<List<T>> list<T, F>([T convert(F d)]) async {
    StringResponse resp = await go;
    if (resp.statusCode == 200) return go.decodeList(convert);
    throw resp;
  }

  AsyncStringResponse expect(List<Checker<Response>> conditions) =>
      go..expect(conditions);

  AsyncStringResponse exact(
          {int statusCode,
          String body,
          List<int> bytes,
          String mimeType,
          String encoding,
          Map<String, String> headers,
          int contentLength}) =>
      go
        ..exact(
            statusCode: statusCode,
            body: body,
            bytes: bytes,
            mimeType: mimeType,
            encoding: encoding,
            headers: headers,
            contentLength: contentLength);
}

/// Build fluent REST PUT APIs
///
/// Example:
///     put('/book/${id}')
///       .json(book.toMap)
///       .fetch((m) => new Book.fromMap(m));
class Put extends Route {
  dynamic _body;

  Put([String authorityOrPath, String path]) : super(authorityOrPath, path);

  Put.copy(Route route) {
    _authority = route._authority;
    _path = route._path;
    _query = route._query;
    _headers = route._headers;
    _client = route._client;
    before.addAll(route.before);
    after.addAll(route.after.toList());
  }

  Put withClient(ht.BaseClient client) => super.withClient(client);

  Put http(String authority, [String path]) => super.http(authority, path);

  Put https(String authority, [String path]) => super.https(authority, path);

  Put authority(String authority, [String path]) =>
      super.authority(authority, path);

  Put path(String path) => super.path(path);

  Put query(String key, value) => super.query(key, value);

  Put queries(Map<String, dynamic> value) => super.queries(value);

  Put header(String key, String value) => super.header(key, value);

  Put headers(Map<String, String> values) => super.headers(values);

  Put cookie(ClientCookie cookie) => super.cookie(cookie);

  Put cookies(List<ClientCookie> cookies) => super.cookies(cookies);

  Put body(String body) {
    _body = body;
    return this;
  }

  Put json(body, {bool setHeaders: true}) {
    _body = codec.json.encode(body);
    if (setHeaders) {
      header('content-type', 'application/json');
      header('Accept', 'application/json');
    }
    return this;
  }

  Put multipart(Map<String, String> values) {
    if (_body is! Map<String, Multipart>) {
      _body = <String, Multipart>{};
    }
    for (String field in values.keys) {
      _body[field] = new _MultipartString(values[field]);
    }
    return this;
  }

  Put multipartFile(String field, List<int> value,
      {String filename, MediaType contentType}) {
    if (_body is! Map<String, Multipart>) {
      _body = <String, Multipart>{};
    }
    _body[field] =
        new _MultipartFile(value, filename: filename, contentType: contentType);
    return this;
  }

  Put multipartStringFile(String field, String value,
      {String filename, MediaType contentType}) {
    if (_body is! Map<String, Multipart>) {
      _body = <String, Multipart>{};
    }
    _body[field] = new _MultipartStringFile(value,
        filename: filename, contentType: contentType);
    return this;
  }

  AsyncStringResponse get go {
    for (Before mod in before) mod(this);
    if (_cookies.length != 0)
      _headers['Cookie'] = ClientCookie.toHeader(_cookies);
    if (_body is String || _body is Map<String, String> || _body == null) {
      return new AsyncStringResponse.from(
          (_client ?? globalClient).put(url, headers: _headers, body: _body));
    } else if (_body is Map<String, Multipart>) {
      final body = _body as Map<String, Multipart>;
      final r = new ht.MultipartRequest('PUT', Uri.parse(url));
      for (final String field in body.keys) {
        final Multipart value = body[field];
        if (value is _MultipartString) {
          r.fields[field] = value.value;
        } else if (value is _MultipartStringFile) {
          r.files.add(new ht.MultipartFile.fromString(field, value.value,
              filename: value.filename, contentType: value.contentType));
        } else if (value is _MultipartFile) {
          r.files.add(new ht.MultipartFile.fromBytes(field, value.value,
              filename: value.filename, contentType: value.contentType));
        }
      }
      r.headers.addAll(_headers);
      return new AsyncStringResponse.from(
          r.send().then((r) => ht.Response.fromStream(r)))
        ..then((r) {
          for (After rp in after) rp(r);
        });
    } else {
      throw new Exception('Invalid body!');
    }
  }

  Future<T> one<T, F>([T convert(F d)]) async {
    StringResponse resp = await go;
    if (resp.statusCode == 200) return go.decode(convert);
    throw resp;
  }

  Future<List<T>> list<T, F>([T convert(F d)]) async {
    StringResponse resp = await go;
    if (resp.statusCode == 200) return go.decodeList(convert);
    throw resp;
  }

  AsyncStringResponse expect(List<Checker<Response>> conditions) =>
      go..expect(conditions);

  AsyncStringResponse exact(
          {int statusCode,
          String body,
          List<int> bytes,
          String mimeType,
          String encoding,
          Map<String, String> headers,
          int contentLength}) =>
      go
        ..exact(
            statusCode: statusCode,
            body: body,
            bytes: bytes,
            mimeType: mimeType,
            encoding: encoding,
            headers: headers,
            contentLength: contentLength);
}

/// Build fluent REST DELETE APIs
///
/// Example:
///     delete('/book/${id}')
///       .fetchList((m) => new Book.fromMap(m));
class Delete extends Route {
  dynamic _body;

  Delete([String authorityOrPath, String path]) : super(authorityOrPath, path);

  Delete.copy(Route route) {
    _authority = route._authority;
    _path = route._path;
    _query = route._query;
    _headers = route._headers;
    _client = route._client;
    before.addAll(route.before);
    after.addAll(route.after.toList());
  }

  Delete withClient(ht.BaseClient client) => super.withClient(client);

  Delete http(String authority, [String path]) => super.http(authority, path);

  Delete https(String authority, [String path]) => super.https(authority, path);

  Delete authority(String authority, [String path]) =>
      super.authority(authority, path);

  Delete path(String path) => super.path(path);

  Delete query(String key, value) => super.query(key, value);

  Delete queries(Map<String, dynamic> value) => super.queries(value);

  Delete header(String key, String value) => super.header(key, value);

  Delete headers(Map<String, String> values) => super.headers(values);

  Delete cookie(ClientCookie cookie) => super.cookie(cookie);

  Delete cookies(List<ClientCookie> cookies) => super.cookies(cookies);

  AsyncStringResponse get go {
    for (Before mod in before) mod(this);
    if (_cookies.length != 0)
      _headers['Cookie'] = ClientCookie.toHeader(_cookies);
    return new AsyncStringResponse.from(
        (_client ?? globalClient).delete(url, headers: _headers))
      ..then((r) {
        for (After rp in after) rp(r);
      });
  }

  Future<T> one<T, F>([T convert(F d)]) async {
    StringResponse resp = await go;
    if (resp.statusCode == 200) return go.decode(convert);
    throw resp;
  }

  Future<List<T>> list<T, F>([T convert(F d)]) async {
    StringResponse resp = await go;
    if (resp.statusCode == 200) return go.decodeList(convert);
    throw resp;
  }

  AsyncStringResponse expect(List<Checker<Response>> conditions) =>
      go..expect(conditions);

  AsyncStringResponse exact(
          {int statusCode,
          String body,
          List<int> bytes,
          String mimeType,
          String encoding,
          Map<String, String> headers,
          int contentLength}) =>
      go
        ..exact(
            statusCode: statusCode,
            body: body,
            bytes: bytes,
            mimeType: mimeType,
            encoding: encoding,
            headers: headers,
            contentLength: contentLength);
}

abstract class Multipart {}

class _MultipartString implements Multipart {
  final String value;

  _MultipartString(this.value);
}

class _MultipartFile implements Multipart {
  final String filename;

  final MediaType contentType;

  final List<int> value;

  _MultipartFile(this.value, {this.filename, this.contentType});
}

class _MultipartStringFile implements Multipart {
  final String filename;

  final MediaType contentType;

  final String value;

  _MultipartStringFile(this.value, {this.filename, this.contentType});
}
